"""
Merkle-tree core functionality
"""

from abc import ABCMeta, abstractmethod

from pymerkle.utils import log2, decompose
from pymerkle.hasher import MerkleHasher
from pymerkle.proof import MerkleProof


class InvalidChallenge(Exception):
    """
    Raised when no Merkle-proof exists for the provided challenge
    """
    pass


class BaseMerkleTree(MerkleHasher, metaclass=ABCMeta):
    """
    Storage agnostic encapsulation of the core Merkle-tree functionality.
    Concrete definitions should inherit from this class and implement its
    private storage interface

    :param algorithm: hash algorithm
    :type algorithm: str
    :param security: [optional] resistance against second-preimage attack.
        Defaults to *True*
    :type security: bool
    """

    @abstractmethod
    def _encode_leaf(self, entry):
        """
        Should return the binary format of the provided entry

        :param entry: data to encode
        :type entry: whatever expected according to application logic
        :rtype: bytes
        """

    @abstractmethod
    def _store_leaf(self, entry, value):
        """
        Should create a new leaf storing the provided entry along with its
        binary format and corresponding hash value

        :param entry: data to append
        :type entry: whatever expected according to application logic
        :param value: hashed data
        :type value: bytes
        :returns: index of newly appended leaf counting from one
        :rtype: int
        """

    @abstractmethod
    def _get_leaf(self, index):
        """
        Should return the hash stored by the leaf specified

        :param index: leaf index counting from one
        :type index: int
        :rtype: bytes
        """

    @abstractmethod
    def _get_size(self):
        """
        Should return the current number of leaves

        :rtype: int
        """

    def append(self, entry):
        """
        Appends a new leaf storing the provided entry

        :param entry: data to append
        :type entry: whatever expected according to application logic
        :returns: index of newly appended leaf counting from one
        :rtype: int
        """
        blob = self._encode_leaf(entry)
        value = self.hash_leaf(blob)
        index = self._store_leaf(entry, value)

        return index


    def get_leaf(self, index):
        """
        Returns the hash of the leaf located at the provided position

        :param index: leaf index counting from one
        :type index: int
        :rtype: bytes
        """
        return self._get_leaf(index)


    def get_size(self):
        """
        Returns the current number of leaves

        :rtype: int
        """
        return self._get_size()


    def get_state(self, subsize=None):
        """
        Computes the root-hash of the subtree corresponding to the provided
        size

        :param subsize: [optional] number of leaves to consider. Defaults to
            current tree size
        :type subsize: int
        :rtype: bytes
        """
        if subsize is None:
            subsize = self._get_size()

        return self.hash_range(0, subsize)


    def hash_range(self, start, end):
        """
        Returns the root-hash corresponding to the provided leaf range

        :param start: first leaf index counting from zero
        :type start: int
        :param end: last leaf index counting from one
        :type end: int
        :rtype: bytes
        """
        if end == start:
            return self.consume(b'')

        if end == start + 1:
            return self.get_leaf(end)

        k = 1 << log2(end - start)
        if k == end - start:
            k >>= 1

        left = self.hash_range(start, start + k)
        right = self.hash_range(start + k, end)

        return self.hash_nodes(left, right)


    def inclusion_path(self, start, offset, end, bit):
        """
        Computes the inclusion path for the leaf located at the provided offset
        against the specified leaf range

        .. warning:: This method should not be called directly. Use
            ``prove_inclusion`` instead

        :param start: leftmost leaf index counting from zero
        :type start: int
        :param offset: base leaf index counting from zero
        :type offset: int
        :param end: rightmost leaf index counting from zero
        :type end: int
        :param bit: indicates direction during recursive call
        :type bit: int
        :rtype: (list[int], list[bytes])
        """
        if offset == start and start == end - 1:
            value = self.get_leaf(offset + 1)
            return [bit], [value]

        k = 1 << log2(end - start)
        if k == end - start:
            k >>= 1

        if offset < start + k:
            rule, path = self.inclusion_path(start, offset, start + k, 0)
            value = self.hash_range(start + k, end)
        else:
            rule, path = self.inclusion_path(start + k, offset, end, 1)
            value = self.hash_range(start, start + k)

        return rule + [bit], path + [value]


    def prove_inclusion(self, index, subsize=None):
        """
        Proves inclusion of the hash located at the provided index against the
        subtree specified by the provided size

        :param index: leaf index counting from one
        :type index: int
        :param subsize: [optional] subsize of subtree to consider. Defaults to
            current tree size
        :type subsize: int
        :rtype: MerkleProof
        :raises InvalidChallenge: if the provided parameters are invalid or
            incompatible with each other
        """
        if subsize is None:
            subsize = self.get_size()

        if subsize > self.get_size():
            raise InvalidChallenge('Provided size is out of bounds')

        if index <= 0 or index > subsize:
            raise InvalidChallenge('Provided index is out of bounds')

        rule, path = self.inclusion_path(0, index - 1, subsize, 0)

        return MerkleProof(self.algorithm, self.security, subsize, rule, [],
                path)


    def consistency_path(self, start, offset, end, bit):
        """
        Computes the consistency path for the state corresponding to the
        provided offset against the specified leaf range

        .. warning:: This method should not be called directly. Use
            ``prove_consistency`` instead

        :param start: leftmost leaf index counting from zero
        :type start: int
        :param offset: size corresponding to state under consideration
        :type offset: int
        :param end: rightmost leaf index counting from zero
        :type end: int
        :param bit: indicates direction during recursive call
        :type bit: int
        :rtype: (list[int], list[int], list[bytes])
        """
        if offset == end:
            value = self.hash_range(start, start + end)
            return [bit], [1], [value]

        if offset == 0 and end == 1:
            value = self.get_leaf(start + offset + 1)
            return [bit], [0], [value]

        k = 1 << log2(end)
        if k == end:
            k >>= 1
        mask = 0

        if offset < k:
            rule, subset, path = self.consistency_path(start, offset, k, 0)
            value = self.hash_range(start + k, start + end)
        else:
            rule, subset, path = self.consistency_path(start + k, offset - k,
                    end - k, 1)
            value = self.hash_range(start, start + k)
            mask = int(k == 1 << log2(k))

        return rule + [bit], subset + [mask], path + [value]


    def prove_consistency(self, size1, size2=None):
        """
        Proves consistency between the states corresponding to the provided
        sizes

        :param size1: size of prior state
        :type size1: int
        :param size2: [optional] size of later state. Defaults to current tree
            size
        :type size2: int
        :rtype: MerkleProof
        :raises InvalidChallenge: if the provided parameters are invalid or
            incompatible with each other
        """
        if size2 is None:
            size2 = self.get_size()

        if size2 < 0 or size2 > self.get_size():
            raise InvalidChallenge('Provided size2 is out of bounds')

        if size1 < 0 or size1 > size2:
            raise InvalidChallenge('Provided size1 is out of bounds')

        rule, subset, path = self.consistency_path(0, size1, size2, 0)

        return MerkleProof(self.algorithm, self.security, size2, rule,
                subset, path)



"""
merkle-tree cryptography
"""

from .concrete.inmemory import InmemoryTree
from .concrete.sqlite import SqliteTree
from .core import BaseMerkleTree, InvalidChallenge
from .proof import MerkleProof, verify_inclusion, verify_consistency, InvalidProof


__version__ = '5.0.3'

__all__ = (
    'BaseMerkleTree',
    'InmemoryTree',
    'SqliteTree',
    'InvalidProof',
    'InvalidChallenge',
    'MerkleProof',
    'verify_inclusion',
    'verify_consistency',
)



def log2(n):
    """
    Base 2 logarithm

    .. note:: This is the exponent of the largest power of 2 which is less than
        or equal to the provided integer

    :param n: non-negative integer
    :type n: int
    :rtype: int
    """
    k = 0
    while n >> 1:
        k += 1
        n >>= 1

    return k


def decompose(n):
    """
    Returns in respective order the exponents corresponding to the binary
    decomposition of the provided integer

    :param n: non-negative integer
    :type n: int
    :rtype: list[int]
    """
    exponents = []

    i = 1
    while i < n + 1:
        if i & n:
            p = -1
            j = i
            while j:
                j >>= 1
                p += 1
            exponents += [p]

        i <<= 1

    return exponents



from pymerkle.utils import decompose
from pymerkle.core import BaseMerkleTree


class Node:
    """
    Merkle-tree node

    :param value: the hash to be stored by the node
    :type value: bytes
    :param left: [optional] left child
    :type left: Node
    :param right: [optional] right child
    :type right: Node
    :rtype: Node
    """

    __slots__ = ('value', 'left', 'right', 'parent')


    def __init__(self, value, left=None, right=None):
        self.value = value

        self.left = left
        if left:
            left.parent = self

        self.right = right
        if right:
            right.parent = self

        self.parent = None


    def is_root(self):
        """
        :rtype: bool
        """
        return not self.parent


    def is_leaf(self):
        """
        :rtype: bool
        """
        return not self.left and not self.right


    def is_left_child(self):
        """
        :rtype: bool
        """
        parent = self.parent
        if not parent:
            return False

        return self == parent.left


    def is_right_child(self):
        """
        :rtype: bool
        """
        parent = self.parent
        if not parent:
            return False

        return self == parent.right


    def get_ancestor(self, degree):
        """
        .. note:: Ancestor of degree 0 is the node itself, ancestor of degree
            1 is the node's parent, etc.

        :type degree: int
        :rtype: Node
        """
        curr = self
        while degree > 0:
            curr = curr.parent
            degree -= 1

        return curr


    def expand(self, indent=2, trim=None, level=0, ignored=None):
        """
        Returns a string representing the subtree rooted at the current node

        :param indent: [optional]
        :type indent: str
        :param trim: [optional]
        :type trim: str
        :param level: [optional]
        :type level: str
        :param ignored: [optional]
        :type ignored: str
        :rtype: str
        """
        ignored = ignored or []

        if level == 0:
            out = 2 * '\n' + ' └─' if not self.parent else ''
        else:
            out = (indent + 1) * ' '

        col = 1
        while col < level:
            out += ' │' if col not in ignored else 2 * ' '
            out += indent * ' '
            col += 1

        if self.is_left_child():
            out += ' ├──'

        if self.is_right_child():
            out += ' └──'
            ignored += [level]

        checksum = self.value.hex()
        out += (checksum[:trim] + '...') if trim else checksum
        out += '\n'

        if self.is_leaf():
            return out

        recursion = (indent, trim, level + 1, ignored[:])

        out += self.left.expand(*recursion)
        out += self.right.expand(*recursion)

        return out


class Leaf(Node):
    """
    Merkle-tree leaf

    :param entry: data stored by the leaf
    :type entry: bytes
    :param value: hash value stored by the leaf
    :type value: bytes
    """

    def __init__(self, entry, value):
        self.entry = entry

        super().__init__(value, None, None)


class InmemoryTree(BaseMerkleTree):
    """
    Non-persistent Merkle-tree with with nodes residing inside the runtime memory

    .. note:: This implementation is intended for debugging and testing. Use it
        to investigate the tree topology by means of concrete path traversals
    """

    def __init__(self, algorithm='sha256', security=True):
        self.root = None
        self.leaves = []

        super().__init__(algorithm, security)


    def __str__(self, indent=2, trim=8):
        """
        :returns: visual representation of the tree
        :rtype: str
        """
        if not self.root:
            return '\n └─[None]\n'

        return self.root.expand(indent, trim) + '\n'


    def _encode_leaf(self, entry):
        """
        Returns the binary format of the provided entry

        :param entry: data to encode
        :type entry: bytes
        :rtype: bytes
        """
        return entry


    def _store_leaf(self, entry, value):
        """
        Creates a new leaf storing the provided entry along with its binary
        format and corresponding hash value

        :param entry: data to append
        :type entry: whatever expected according to application logic
        :param value: hashed data
        :type value: bytes
        :returns: index of newly appended leaf counting from one
        :rtype: int
        """
        tail = Leaf(entry, value)

        if not self.leaves:
            self.leaves += [tail]
            self.root = tail
            return 1

        node = self.get_last_maximal_perfect_node()
        self.leaves += [tail]
        value = self.hash_nodes(node.value, tail.value)

        if node.is_root():
            self.root = Node(value, node, tail)
            index = self.get_size()
            return index

        curr = node.parent
        curr.right = Node(value, node, tail)
        curr.right.parent = curr
        while curr:
            curr.value = self.hash_nodes(
                curr.left.value, curr.right.value)
            curr = curr.parent

        index = self.get_size()
        return index


    def _get_leaf(self, index):
        """
        Returns the hash stored by the leaf specified

        :param index: leaf index counting from one
        :type index: int
        :rtype: bytes
        """
        if index < 1 or index > len(self.leaves):
            raise ValueError("%d not in leaf range" % index)

        return self.leaves[index - 1].value


    def _get_size(self):
        """
        :returns: current number of leaves
        :rtype: int
        """
        return len(self.leaves)


    @classmethod
    def init_from_entries(cls, entries, algorithm='sha256', security=True):
        """
        Create tree from initial data

        :param entries: initial data to append
        :type entries: iterable of bytes
        :param algorithm: [optional] hash function. Defaults to *sha256*
        :type algorithm: str
        :param security: [optional] resistance against second-preimage attack.
            Defaults to *True*
        :type security: bool
        """
        tree = cls(algorithm, security)

        append = tree.append
        for entry in entries:
            append(entry)

        return tree


    def get_state(self, subsize=None):
        """
        Computes the root-hash of the subtree corresponding to the provided
        size

        .. note:: Overrides the default implementation inherited from the base
            class

        :param subsize: [optional] number of leaves to consider. Defaults to
            current tree size
        :type subsize: int
        :rtype: bytes
        """
        if subsize is None:
            subsize = self.get_size()

        if subsize == 0:
            return self.consume(b'')

        if subsize == self.get_size():
            return self.root.value

        principals = self.get_principal_nodes(subsize)
        result = principals[0].value
        i = 0
        while i < len(principals) - 1:
            result = self.hash_nodes(principals[i + 1].value, result)
            i += 1

        return result


    def inclusion_path_fallback(self, offset):
        """
        Non-recursive utility using concrete traversals to compute the inclusion
        path of the provided leaf hash against the current tree state

        .. warning:: This method is intended for investigating the tree
            structure and testing. Use ``prove_inclusion`` to properly generate
            inclusion proofs

        :param offset: base leaf index counting from zero
        :type offset: int
        :rtype: (list[int], list[bytes])
        """
        base = self.leaves[offset]
        bit = 1 if base.is_right_child() else 0

        path = [base.value]
        rule = [bit]

        curr = base
        while curr.parent:
            parent = curr.parent

            if curr.is_left_child():
                value = parent.right.value
                bit = 0 if parent.is_left_child() else 1
            else:
                value = parent.left.value
                bit = 1 if parent.is_right_child() else 0

            rule += [bit]
            path += [value]
            curr = parent

        # Last bit is insignificant; fix it to zero just to be fully compatible
        # with the output of the overriden method
        rule[-1] = 0

        return rule, path


    def inclusion_path(self, start, offset, end, bit):
        """
        Computes the inclusion path for the leaf located at the provided offset
        against the specified leaf range

        .. warning:: This method should not be called directly. Use
            ``prove_inclusion`` instead

        :param start: leftmost leaf index counting from zero
        :type start: int
        :param offset: base leaf index counting from zero
        :type offset: int
        :param end: rightmost leaf index counting from zero
        :type end: int
        :param bit: bit indicating direction during recursive call
        :type bit: int
        :rtype: (list[int], list[bytes])
        """
        if start == 0 and end == self.get_size():
            return self.inclusion_path_fallback(offset)

        return super().inclusion_path(start, offset, end, bit)


    def get_perfect_node(self, index, height):
        """
        Returns the root node of the perfect subtree of the provided height whose
        leftmost leaf node is located at the provided position

        .. note:: Returns *None* if no binary subtree exists for the provided
            parameters

        :param index: position of leftmost leaf node coutning from one
        :type index: int
        :param height: height of requested subtree
        :type height: int
        :rtype: Node
        """
        node = self.leaves[index - 1]

        if not node:
            return

        i = 0
        while i < height:
            curr = node.parent

            if not curr:
                return

            if curr.left is not node:
                return

            node = curr
            i += 1

        # Verify existence of perfect subtree rooted at the detected node
        curr = node
        i = 0
        while i < height:
            if curr.is_leaf():
                return

            curr = curr.right
            i += 1

        return node


    def get_last_maximal_perfect_node(self):
        """
        Returns the root-node of the perfect subtree of maximum possible size
        containing the currently last leaf

        :rtype: Node
        """
        degree = decompose(len(self.leaves))[0]

        return self.leaves[-1].get_ancestor(degree)


    def get_principal_nodes(self, subsize):
        """
        Returns in respective order the root-nodes of the successive perfect
        subtrees whose sizes sum up to the provided size

        :param subsize:
        :type subsize: int
        :rtype: list[Node]
        """
        if subsize < 0 or subsize > self.get_size():
            return []

        principals = []
        offset = 0
        for height in reversed(decompose(subsize)):
            node = self.get_perfect_node(offset + 1, height)

            if not node:
                return []

            principals += [node]
            offset += 1 << height

        return list(reversed(principals))



import hashlib
from pymerkle import constants


class UnsupportedParameter(Exception):
    """
    Raised when a Merkle-hasher with unsupported parameters is requested
    """
    pass


class MerkleHasher:
    """
    Encapsulates elementary hashing operations

    :param algorithm: hash algorithm
    :type algorithm: str
    :param security: [optional] defense against second-preimage attack. Defaults
        to *True*
    :type security: bool
    """

    def __init__(self, algorithm, security=True, **kw):
        normalized = algorithm.lower().replace('-', '_')
        if normalized not in constants.ALGORITHMS:
            raise UnsupportedParameter('%s is not supported' % algorithm)

        self.algorithm = algorithm
        self.security = security
        self.prefx00 = b'\x00' if security else b''
        self.prefx01 = b'\x01' if security else b''


    def consume(self, buff):
        """
        Computes the raw hash of the provided data

        :param buff:
        :type buff: bytes
        :rtype: bytes
        """
        _hasher = getattr(hashlib, self.algorithm)()

        update = _hasher.update
        chunksize = 1024
        offset = 0
        chunk = buff[offset: chunksize]
        while chunk:
            update(chunk)
            offset += chunksize
            chunk = buff[offset: offset + chunksize]

        return _hasher.digest()


    def hash_leaf(self, blob):
        """
        Computes the hash of the provided binary data

        .. note:: Prepends ``\\x00`` if security mode is enabled

        :type blob: bytes
        :rtype: bytes
        """
        buff = self.prefx00 + blob

        return self.consume(buff)


    def hash_nodes(self, blob1, blob2):
        """
        Computes the hash of the concatenation of the provided binary data

        .. note:: Prepends ``\\x01`` if security mode is enabled

        :param blob1: left value
        :type blob1: bytes
        :param blob2: right value
        :type blob2: bytes
        :rtype: bytes
        """
        buff = self.prefx01 + blob1 + blob2

        return self.consume(buff)



import sqlite3
from pymerkle.core import BaseMerkleTree


class SqliteTree(BaseMerkleTree):
    """
    Persistent Merkle-tree implementation using a SQLite database as storage

    The database schema consists of a single table called *leaf* with two
    columns: *index*, which is the primary key serving as leaf index, and
    *entry*, which is a blob field storing the appended data. Inserted data are
    expected by the tree to be in binary format and stored without further
    processing

    :param dbfile: database filepath
    :type dbfile: str
    :param algorithm: [optional] hashing algorithm. Defaults to *sha256*
    :type algorithm: str
    :param security: [optional] resistance against second-preimage attack.
        Defaults to *True*
    :type security: bool
    """

    def __init__(self, dbfile, algorithm='sha256', security=True):
        self.con = sqlite3.connect(dbfile)
        self.cur = self.con.cursor()

        with self.con:
            query = f'''
                CREATE TABLE IF NOT EXISTS leaf(
                    id INTEGER PRIMARY KEY AUTOINCREMENT,
                    entry BLOB,
                    hash BLOB
                );'''
            self.cur.execute(query)

        super().__init__(algorithm, security)


    def __enter__(self):
        return self


    def __exit__(self, *exc):
        self.con.close()


    def _encode_leaf(self, entry):
        """
        Returns the binary format of the provided entry

        :param entry: data to encode
        :type entry: bytes
        :rtype: bytes
        """
        return entry


    def _store_leaf(self, entry, value):
        """
        Creates a new leaf storing the provided entry along with its binary
        format and corresponding hash value

        :param entry: data to append
        :type entry: whatever expected according to application logic
        :param value: hashed data
        :type value: bytes
        :returns: index of newly appended leaf counting from one
        :rtype: int
        """
        if not isinstance(entry, bytes):
            raise ValueError('Provided data is not binary')

        cur = self.cur

        with self.con:
            query = f'''
                INSERT INTO leaf(entry, hash) VALUES (?, ?)
            '''
            cur.execute(query, (entry, value))

        return cur.lastrowid

    def _get_leaf(self, index):
        """
        Returns the hash stored by the leaf specified

        :param index: leaf index counting from one
        :type index: int
        :rtype: bytes
        """
        cur = self.cur

        query = f'''
            SELECT hash FROM leaf WHERE id = ?
        '''
        cur.execute(query, (index,))

        return cur.fetchone()[0]


    def _get_size(self):
        """
        :returns: current number of leaves
        :rtype: int
        """
        cur = self.cur

        query = f'''
            SELECT COUNT(*) FROM leaf
        '''
        cur.execute(query)

        return cur.fetchone()[0]


    def get_entry(self, index):
        """
        Returns the original data stored by the leaf specified

        :param index: leaf index counting from one
        :type index: int
        :rtype: bytes
        """
        cur = self.cur

        query = f'''
            SELECT entry FROM leaf WHERE id = ?
        '''
        cur.execute(query, (index,))

        return cur.fetchone()[0]



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

from setuptools import setup, find_packages
import os
import io

import pymerkle


URL = "https://github.com/fmerg/pymerkle"

current_dir = os.path.abspath(os.path.dirname(__file__))

try:
  with io.open(os.path.join(current_dir, "requirements.txt"),
    encoding="utf-8") as f:
    install_requires = [_.strip() for _ in f.readlines()]
except FileNotFoundError:
    install_requires = []

with open("README.md", 'r') as f:
    long_description = f.read()

def main():
    setup(
       name=pymerkle.__name__,
       version=pymerkle.__version__,
       description=pymerkle.__doc__.strip(),
       long_description=long_description,
       long_description_content_type='text/markdown',
       packages=find_packages(exclude=['tests']),
       # package_dir={'': 'pymerkle'},
       url=URL,
       project_urls={
            "github": URL,
            "source": "%s/%s" % (URL, "tree/master/%s" % pymerkle.__name__),
            "docs": "https://%s.readthedocs.io/en/latest/" % pymerkle.__name__,
       },
       author="fmerg",
       author_email="fmerg@protonmail.com",
       python_requires=">=3.6",
       install_requires=install_requires,
       zip_safe=False,
       keywords=[
           "merkle", "proof", "inclusion", "consistency",
       ],
       classifiers=[
           "Development Status :: 4 - Beta",
           "Intended Audience :: Developers",
           "Intended Audience :: Science/Research",
           "Programming Language :: Python :: 3.6",
           "Operating System :: POSIX",
           "Topic :: Security :: Cryptography",
           "Topic :: Software Development :: Libraries :: Python Modules",
           "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)"
       ],
    )


if __name__ == '__main__':
    main()



import os
import json
from time import time

from pymerkle.hasher import MerkleHasher


class InvalidProof(Exception):
    """
    Raised when a Merkle-proof is found to be invalid
    """
    pass


def verify_inclusion(base, target, proof):
    """
    Verifies the provided Merkle-proof of inclusion against the provided base
    hash and state

    :param base: acclaimed base hash
    :type base: bytes
    :param target: acclaimed root hash
    :type target: bytes
    :param proof: proof of inclusion
    :type proof: MerkleProof
    :raises InvalidProof: if the proof is found invalid
    """
    if not proof.path[0] == base:
        raise InvalidProof('Base hash does not match')

    if not proof.resolve() == target:
        raise InvalidProof('State does not match')


def verify_consistency(state1, state2, proof):
    """
    Verifies the provided Merkle-proof of consistency against the given states

    :param state1: acclaimed prior state
    :type state1: bytes
    :param state2: acclaimed later state
    :type state2: bytes
    :raises InvalidProof: if the proof is found invalid
    :param proof: proof of consistency
    :type proof: MerkleProof
    """
    if not proof.retrieve_prior_state() == state1:
        raise InvalidProof('Prior state does not match')

    if not proof.resolve() == state2:
        raise InvalidProof('Later state does not match')


class MerkleProof(MerkleHasher):
    """
    Verifiable Merkle-proof object

    :param algorithm: hash algorithm
    :type algorithm: str
    :param security: resistance against 2-nd preimage attack
    :type security: bool
    :param size: tree size corresponding to requested state
    :type size: int
    :param rule: specifies parenthetization of hashes during state
        resolution
    :type rule: list[int]
    :param subset: indicates subset of hashes for prior state resolution
        (makes sense only for proofs of consistency)
    :type subset: list[int]
    :param path: path of hashes
    :type path: list[bytes]
    """

    def __init__(self, algorithm, security, size, rule, subset, path):
        self.algorithm = algorithm
        self.security = security
        self.size = size
        self.rule = rule
        self.subset = subset
        self.path = path

        MerkleHasher.__init__(self, **self.get_metadata())


    def get_metadata(self):
        """
        :rtype: dict
        """
        return {'algorithm': self.algorithm, 'security': self.security,
                'size': self.size}

    def serialize(self):
        """
        :rtype: dict
        """
        return {
            'metadata': {
                'algorithm': self.algorithm,
                'security': self.security,
                'size': self.size,
            },
            'rule': self.rule,
            'subset': self.subset,
            'path': [value.hex() for value in self.path]
        }


    @classmethod
    def deserialize(cls, data):
        """
        :param data:
        :type data: dict
        :rtype: MerkleProof
        """
        metadata = data['metadata']
        rule = data['rule']
        subset = data['subset']
        path = [bytes.fromhex(value) for value in data['path']]

        return cls(**metadata, rule=rule, subset=subset, path=path)


    def retrieve_prior_state(self):
        """
        Computes the acclaimed prior state which is immanent in the included
        path of hashes

        .. note:: Makes sense only for consistency proofs

        :rtype: bytes
        """
        subpath = [value for (mask, value) in zip(self.subset, self.path) if
            mask]

        if not subpath:
            return self.consume(b'')

        result = subpath[0]
        index = 0
        while index < len(subpath) - 1:
            result = self.hash_nodes(subpath[index + 1], result)
            index += 1

        return result


    def resolve(self):
        """
        Computes the target of the included path of hashes

        :rtype: bytes
        """
        path = list(zip(self.rule, self.path))

        if not path:
            return self.consume(b'')

        bit, result = path[0]
        index = 0
        while index < len(path) - 1:
            next_bit, value = path[index + 1]

            match bit:
                case 0:
                    result = self.hash_nodes(result, value)
                case 1:
                    result = self.hash_nodes(value, result)
                case _:
                    raise Exception('Invalid bit found')

            bit = next_bit
            index += 1

        return result
